#include <stdio.h>
#include <math.h>
#include <float.h>
#ifdef USE_MPI
# include <mpi.h>
#endif
#include "hrtimer.h"

#define LOGISTIC_MAP(r, x)  (r)*(x)*(1-(x))
#define INF_LOGISTIC_MAP(x) (3.6*(x))
#define SIN_MAP(x)	    (INF_LOGISTIC_MAP(sin((x))))

#define USEC_WARMUP 1000L
#define USEC_TIMING 10000000L
HRTIME_T cycles_per_usec;

/* logarithmically sized bins, 100 bins per decade, time converted to nsec */
#define NBINS		1500
/* bins[NBINS] is underflow and bins[NBINS+1] is overflow */
#define NXBINS		(NBINS+2)
#define BIN(x)		(int) (log10((x)*1000/cycles_per_usec)*100.0)
#define USEC(bin)	( (exp10((bin+1)/100.0) + exp10((bin+1)/100.0))/2.0/1000.0 )

double
loop(unsigned long usecs, unsigned long *hbin)
{
    double x = 0.3;
    int j;
    HRTIME_T hb, h1, h2, hd, maxcycles, iter = 0, sum = 0;

    if (hbin)
	for (j = 0; j < NXBINS; j++)
	    hbin[j] = 0;
    cycles_per_usec = HRTIMER_USEC();
    maxcycles = usecs*cycles_per_usec;
    h2 = hb = HRTIMER();
    while ((h2 - hb) < maxcycles) {
	h1 = HRTIMER();
	x = SIN_MAP(x);
	h2 = HRTIMER();
	hd = h2 - h1;
	sum += hd;
	iter++;
	j = BIN(hd);
	if (hbin)
	    if (j < 0)
		hbin[NBINS]++;
	    else if (j >= NBINS)
		hbin[NBINS+1]++;
	    else
		hbin[j]++;
    }
    return sum*1.0/iter/cycles_per_usec;
}

int
fprintbins(FILE *fp, unsigned long *hbin)
{
    int i;

    if (hbin[NBINS] > 0 || hbin[NBINS+1] > 0)
	fprintf(fp, "# underflow %lu overflow %lu\n", hbin[NBINS], hbin[NBINS+1]);
    for (i = 0; i < NBINS; i++) {
	if (hbin[i] > 0 || (i > 0 && hbin[i-1] != 0)) {
	    fprintf(fp, "%g %lu\n", USEC(i), hbin[i]);
	}
    }
    fprintf(fp, "\n\n");
}

int
mpi_main(int argc, char **argv)
{
#ifdef USE_MPI
    unsigned long mybins[NXBINS], sumbins[NXBINS];
    int i, rank, nprocs;
    double d, *sumd;

    for (i = 0; i < NXBINS; i++)
	sumbins[i] = 0;
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (rank == 0) {
	sumd = calloc(nprocs, sizeof(double));
	if (sumd == NULL) {
	    perror("logmap: calloc");
	    exit(2);
	}
    }
    (void) loop(USEC_WARMUP, NULL);
    d = loop(USEC_TIMING, mybins);
    MPI_Reduce(mybins, sumbins, NXBINS, MPI_UNSIGNED_LONG, MPI_SUM, 0,
	       MPI_COMM_WORLD);
    MPI_Gather(&d, 1, MPI_DOUBLE, sumd, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);  
    d = 0;
    if (rank == 0) {
	double dmin = DBL_MAX, dmax = DBL_MIN;
	for (i = 0; i < nprocs; i++) {
	    d += sumd[i];
	    if (sumd[i] > dmax)
		dmax = sumd[i];
	    if (sumd[i] < dmin)
		dmin = sumd[i];
	}
	printf("# %d procs, min/avg/max avg = %g / %g / %g\n", nprocs, dmin, d/nprocs, dmax);
	fprintbins(stdout,sumbins);
    }
    MPI_Finalize();
#else
    printf("MPI not supported\n");
#endif
    return 0;
}


int
normal_main(int argc, char **argv)
{
    unsigned long hbin[NXBINS];
    double d;

    (void) loop(USEC_WARMUP, NULL);
    d = loop(USEC_TIMING, hbin);
    printf("# avg = %g\n", d);
    fprintbins(stdout, hbin);
    return 0;
}

int
main(int argc, char **argv)
{
    if (argc > 1 && strcmp(argv[1], "mpi") == 0)
	mpi_main(argc, argv);
    else
	normal_main(argc, argv);
    return 0;
}
